from typing import Annotated, Self

from pydantic import BaseModel, ConfigDict, Field, model_validator

from lomas_core.constants import (
    DPLibraries,
    SSynthGanSynthesizer,
    SSynthMarginalSynthesizer,
)
from lomas_core.error_handler import InternalServerException
from lomas_core.models.constants import JSON_SCHEMA_EXAMPLES, PrivateDatabaseType
from lomas_core.models.requests_examples import (
    example_diffprivlib,
    example_dummy_diffprivlib,
    example_dummy_opendp,
    example_dummy_smartnoise_sql,
    example_dummy_smartnoise_synth_query,
    example_opendp_polars,
    example_smartnoise_sql,
    example_smartnoise_sql_cost,
    example_smartnoise_synth_cost,
    example_smartnoise_synth_query,
)


class LomasRequestModel(BaseModel):
    """Base class for all types of requests to the lomas server.

    We differentiate between requests and queries:
        - a request does not necessarily require an algorithm
          to be executed on the private dataset (e.g. some cost requests).
        - a query requires executing an algorithm on a private
          dataset (or a potentially a dummy).
    """

    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset_name: str
    """The name of the dataset the request is aimed at."""


class AddDatasetModel(LomasRequestModel):
    """Model input to add a private dataset with metadata."""

    database_type: PrivateDatabaseType
    """Type of Private Database for the private data."""
    metadata_database_type: PrivateDatabaseType
    """Type of Private Database for the private data."""
    dataset_path: str
    """Path to the dataset."""
    metadata_path: str
    """Path to the metadata."""


class GetDummyDataset(LomasRequestModel):
    """Model input to get a dummy dataset."""

    dummy_nb_rows: Annotated[int, Field(gt=0)]
    """The number of dummy rows to generate."""
    dummy_seed: int
    """The seed for the random generation of the dummy dataset."""


class GetDummyContext(GetDummyDataset):
    """Model input to get a dummy dataset."""

    epsilon: Annotated[float | None, Field(ge=0.0)]
    """The epsilon parameter used for pure ε-DP or approximate-DP."""
    delta: Annotated[float | None, Field(ge=0.0)]
    """The delta parameter."""
    rho: Annotated[float | None, Field(ge=0.0)]
    """
    Privacy loss paramater for zCDP (or approximate-zCDP).

    Using this parameter instead of `epsilon` switches to a Gaussian mechansim.
    """
    approx_zcdp: bool
    """If False, delta is used to compute the epsilon consumption equivalent when user wants to use zCDP."""


class QueryModel(LomasRequestModel):
    """
    Base input model for any query on a dataset.

    We differentiate between requests and queries:
        - a request does not necessarily require an algorithm
          to be executed on the private dataset (e.g. some cost requests).
        - a query requires executing an algorithm on a private
          dataset (or a potentially a dummy).
    """


class LomasBudgetRequest(LomasRequestModel):
    epsilon: Annotated[float, Field(gt=0)]
    """Privacy parameter (e.g., 0.1)."""
    delta: Annotated[float, Field(ge=0)]
    """Privacy parameter (e.g., 1e-5)."""


class DummyQueryModel(QueryModel):
    """Input model for a query on a dummy dataset."""

    dummy_nb_rows: Annotated[int, Field(gt=0)]
    """The number of rows in the dummy dataset."""
    dummy_seed: int
    """The seed to set at the start of the dummy dataset generation."""


# SmartnoiseSQL
# ----------------------------------------------------------------------------
class SmartnoiseSQLRequestModel(LomasRequestModel):
    """Base input model for a smarnoise-sql request."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_smartnoise_sql_cost]})

    query_str: str
    """The SQL query to execute.

    NOTE: the table name is \"df\", the query must end with \"FROM df\"
    """
    epsilon: Annotated[float, Field(gt=0)]
    """Privacy parameter (e.g., 0.1)."""
    delta: Annotated[float, Field(ge=0)]
    """Privacy parameter (e.g., 1e-5)."""
    mechanisms: dict
    """
    Dictionary of mechanisms for the query.

    See Smartnoise-SQL mechanisms documentation at
    https://docs.smartnoise.org/sql/advanced.html#overriding-mechanisms.
    """


class SmartnoiseSQLQueryModel(SmartnoiseSQLRequestModel, QueryModel):
    """Base input model for a smartnoise-sql query."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_smartnoise_sql]})

    postprocess: bool
    """
    Whether to postprocess the query results (default: True).

    See Smartnoise-SQL postprocessing documentation
    https://docs.smartnoise.org/sql/advanced.html#postprocess.
    """


class SmartnoiseSQLDummyQueryModel(SmartnoiseSQLQueryModel, DummyQueryModel):
    """Input model for a smartnoise-sql query on a dummy dataset."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_dummy_smartnoise_sql]})


# SmartnoiseSynth
# ----------------------------------------------------------------------------
class SmartnoiseSynthRequestModel(LomasRequestModel):
    """Base input model for a SmartnoiseSynth request."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_smartnoise_synth_cost]})

    synth_name: SSynthMarginalSynthesizer | SSynthGanSynthesizer
    """Name of the synthesizer model to use."""
    epsilon: Annotated[float, Field(gt=0)]
    """Privacy parameter (e.g., 0.1)."""
    delta: Annotated[float | None, Field(ge=0)]
    """Privacy parameter (e.g., 1e-5)."""
    select_cols: list
    """List of columns to select."""
    synth_params: dict
    """
    Keyword arguments to pass to the synthesizer constructor.

    See https://docs.smartnoise.org/synth/synthesizers/index.html#, provide
    all parameters of the model except `epsilon` and `delta`.
    """
    nullable: bool
    """True if some data cells may be null."""
    constraints: str
    """
    Dictionnary for custom table transformer constraints.

    Column that are not specified will be inferred based on metadata.
    """


class SmartnoiseSynthQueryModel(SmartnoiseSynthRequestModel, QueryModel):
    """Base input model for a smarnoise-synth query."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_smartnoise_synth_query]})

    return_model: bool
    """True to get Synthesizer model, False to get samples."""
    condition: str
    """Sampling condition in `model.sample` (only relevant if return_model is False)."""
    nb_samples: int
    """Number of samples to generate.

    (only relevant if return_model is False)
    """


class SmartnoiseSynthDummyQueryModel(SmartnoiseSynthQueryModel, DummyQueryModel):
    """Input model for a smarnoise-synth query on a dummy dataset."""

    model_config = ConfigDict(
        json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_dummy_smartnoise_synth_query]}
    )


# OpenDP
# ----------------------------------------------------------------------------
class OpenDPRequestModel(LomasRequestModel):
    """Base input model for an opendp request."""

    model_config = ConfigDict(
        use_attribute_docstrings=True,
        json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_opendp_polars]},
    )

    opendp_json: str
    """The OpenDP pipeline for the query."""
    epsilon: Annotated[float | None, Field(ge=0)]
    """The epsilon parameter used for pure ε-DP or approximate-DP."""
    delta: Annotated[float | None, Field(ge=0)]
    """
    If the pipeline measurement is of type "ZeroConcentratedDivergence".

    (e.g. with "make_gaussian") then it is converted to "SmoothedMaxDivergence"
    with "make_zCDP_to_approxDP" (see "opendp measurements documentation at
    https://docs.opendp.org/en/stable/api/python/opendp.combinators.html#opendp.combinators.make_zCDP_to_approxDP).
    In that case a "delta" must be provided by the user.
    """
    rho: Annotated[float | None, Field(ge=0)]
    """
    Privacy loss parameter for zCDP (or approximate zCDP).

    Using this parameter instead of `epsilon` switches to a Gaussian mechansim.
    """

    approx_zcdp: bool
    """If false, delta is used to compute the epsilon consumption equivalent when user wants to use zCDP."""

    @model_validator(mode="after")
    def check_epsilon_or_rho(self) -> Self:
        if (self.epsilon is None and self.rho is None) or (self.epsilon and self.rho):
            raise ValueError("Either `epsilon` or `rho` must be set.")
        return self


class OpenDPQueryModel(OpenDPRequestModel, QueryModel):
    """Base input model for an opendp query."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_opendp_polars]})


class OpenDPDummyQueryModel(OpenDPRequestModel, DummyQueryModel):
    """Input model for an opendp query on a dummy dataset."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_dummy_opendp]})


# DiffPrivLib
# ----------------------------------------------------------------------------
class DiffPrivLibRequestModel(LomasRequestModel):
    """Base input model for a diffprivlib request."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_diffprivlib]})

    diffprivlib_json: str
    """The DiffPrivLib pipeline for the query (See diffprivlib_logger package.)."""
    feature_columns: list
    """The list of feature columns to train."""
    target_columns: list | None
    """The list of target columns to predict."""
    test_size: Annotated[float, Field(gt=0.0, lt=1.0)]
    """The proportion of the test set."""
    test_train_split_seed: int
    """The seed for the random train/test split."""
    imputer_strategy: str
    """The imputation strategy."""


class DiffPrivLibQueryModel(DiffPrivLibRequestModel, QueryModel):
    """Base input model for a diffprivlib query."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_diffprivlib]})


class DiffPrivLibDummyQueryModel(DiffPrivLibQueryModel, DummyQueryModel):
    """Input model for a DiffPrivLib query on a dummy dataset."""

    model_config = ConfigDict(json_schema_extra={JSON_SCHEMA_EXAMPLES: [example_dummy_diffprivlib]})


# Utils
# ----------------------------------------------------------------------------


def model_input_to_lib(request: LomasRequestModel) -> DPLibraries:
    """Return the type of DP library given a LomasRequestModel.

    Args:
        request (LomasRequestModel): The user request

    Raises:
        InternalServerException: If the library type cannot be determined.

    Returns:
        DPLibraries: The type of library for the request.
    """
    match request:
        case SmartnoiseSQLRequestModel():
            return DPLibraries.SMARTNOISE_SQL
        case SmartnoiseSynthRequestModel():
            return DPLibraries.SMARTNOISE_SYNTH
        case OpenDPRequestModel():
            return DPLibraries.OPENDP
        case DiffPrivLibRequestModel():
            return DPLibraries.DIFFPRIVLIB
        case _:
            raise InternalServerException("Cannot find library type for given model.")



from typing import Any, Never

from fastapi import FastAPI, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from lomas_core.constants import DPLibraries
from lomas_core.models.constants import init_logging
from lomas_core.models.exceptions import (
    ExternalLibraryExceptionModel,
    InternalServerExceptionModel,
    InvalidQueryExceptionModel,
    LomasServerExceptionType,
    UnauthorizedAccessExceptionModel,
)

logger = init_logging(__name__)


class InvalidQueryException(Exception):
    """
    Custom exception for invalid queries.

    For example, this exception will occur when the query:
        - is not an opendp measurement
        - cannot be reconstructed properly (for opendp and diffprivlib)
    """

    def __init__(self, error_message: str) -> None:
        """Invalid Query Exception initialisation.

        Args:
            error_message (str): _description_
        """
        self.error_message = error_message


class ExternalLibraryException(Exception):
    """
    Custom exception for issues within external libraries.

    This exception will occur when the processes fail within the
    external libraries (smartnoise-sql, opendp, diffprivlib)
    """

    def __init__(self, library: DPLibraries, error_message: str) -> None:
        """External Query Exception initialisation.

        Args:
            library (str): _description_
            error_message (str): _description_
        """
        self.library = library
        self.error_message = error_message


class UnauthorizedAccessException(Exception):
    """
    Custom exception for unauthorized access:

    (unknown user, no access to dataset, etc)
    """

    def __init__(self, error_message: str) -> None:
        self.error_message = error_message


class InternalServerException(Exception):
    """Custom exception for issues within server internal functionalities."""

    def __init__(self, error_message: str) -> None:
        self.error_message = error_message


KNOWN_EXCEPTIONS: tuple[type[BaseException], ...] = (
    ExternalLibraryException,
    InternalServerException,
    InvalidQueryException,
    UnauthorizedAccessException,
)


# Custom exception handlers
def add_exception_handlers(app: FastAPI) -> None:
    """
    Translates custom exceptions to JSONResponses.

    Args:
        app (FastAPI): A fastapi App.
    """

    @app.exception_handler(InvalidQueryException)
    async def invalid_query_exception_handler(_: Request, exc: InvalidQueryException) -> JSONResponse:
        logger.debug(f"InvalidQueryException raised: {exc.error_message}")
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content=jsonable_encoder(InvalidQueryExceptionModel(message=exc.error_message)),
        )

    @app.exception_handler(ExternalLibraryException)
    async def external_library_exception_handler(_: Request, exc: ExternalLibraryException) -> JSONResponse:
        logger.debug(f"ExternalLibraryException raised: {exc.error_message}")
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_CONTENT,
            content=jsonable_encoder(
                ExternalLibraryExceptionModel(message=exc.error_message, library=exc.library)
            ),
        )

    @app.exception_handler(UnauthorizedAccessException)
    async def unauthorized_access_exception_handler(
        _: Request, exc: UnauthorizedAccessException
    ) -> JSONResponse:
        logger.debug(f"UnauthorizedAccessException raised: {exc.error_message}")
        return JSONResponse(
            status_code=status.HTTP_403_FORBIDDEN,
            content=jsonable_encoder(UnauthorizedAccessExceptionModel(message=exc.error_message)),
        )

    @app.exception_handler(InternalServerException)
    async def internal_server_exception_handler(_: Request, exc: InternalServerException) -> JSONResponse:
        logger.debug(f"InternalServerException  raised: {exc.error_message}")
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content=jsonable_encoder(InternalServerExceptionModel()),
        )


# Server error responses for DP queries
SERVER_QUERY_ERROR_RESPONSES: dict[int | str, dict[str, Any]] = {
    status.HTTP_400_BAD_REQUEST: {"model": InvalidQueryExceptionModel},
    status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ExternalLibraryExceptionModel},
    status.HTTP_403_FORBIDDEN: {"model": UnauthorizedAccessExceptionModel},
    status.HTTP_500_INTERNAL_SERVER_ERROR: {"model": InternalServerExceptionModel},
}


def raise_error_from_model(error_model: LomasServerExceptionType) -> Never:
    """Raise error message based on Server Error Model.

    Args:
        error_model
    Raise:
        Server Error
    """
    match error_model:
        case InvalidQueryExceptionModel():
            raise InvalidQueryException(error_model.message)
        case ExternalLibraryExceptionModel():
            raise ExternalLibraryException(error_model.library, error_model.message)
        case UnauthorizedAccessExceptionModel():
            raise UnauthorizedAccessException(error_model.message)
        case InternalServerExceptionModel():
            raise InternalServerException("Internal Server Exception.")



from enum import StrEnum

# Server error messages
INTERNAL_SERVER_ERROR = "Internal server error. Please contact the administrator of this service."

TRACE_LOG_LEVEL = 5


class DPLibraries(StrEnum):
    """Name of DP Library used in the query."""

    SMARTNOISE_SQL = "smartnoise_sql"
    SMARTNOISE_SYNTH = "smartnoise_synth"
    OPENDP = "opendp"
    OPENDP_POLARS = "opendp_polars"
    DIFFPRIVLIB = "diffprivlib"


# OpenDP


# Smartnoise synth
class SSynthMarginalSynthesizer(StrEnum):
    """Marginal Synthesizer models for smartnoise synth."""

    AIM = "aim"
    MWEM = "mwem"
    MST = "mst"
    PAC_SYNTH = "pacsynth"


class SSynthGanSynthesizer(StrEnum):
    """GAN Synthesizer models for smartnoise synth."""

    DP_CTGAN = "dpctgan"
    PATE_CTGAN = "patectgan"
    PATE_GAN = "pategan"
    DP_GAN = "dpgan"


# Security
# ------------------------------------------------


class Scopes(StrEnum):
    """List of security scopes for the server endpoints."""

    ADMIN = "admin"


OIDC_LOMAS_CLIENT__CLIENT_ID = "lomas_client"



import logging
from collections.abc import Sequence
from enum import IntEnum, StrEnum
from importlib import metadata
from typing import Literal

from rich.logging import RichHandler

# Field names
# -----------------------------------------------------------------------------

DB_TYPE_FIELD = "database_type"
TYPE_FIELD = "type"
CARDINALITY_FIELD = "cardinality"

JSON_SCHEMA_EXAMPLES = "examples"


# Requests
# -----------------------------------------------------------------------------

DUMMY_NB_ROWS = 100
DUMMY_SEED = 42

OPENDP_VERSION = metadata.version("opendp")
OpenDPFeatures = Sequence[Literal["contrib", "floating-point", "honest-but-curious"]]
DIFFPRIVLIB_VERSION = metadata.version("diffprivlib-fix")


# Metadata
# -----------------------------------------------------------------------------


class MetadataColumnType(StrEnum):
    """Column types for metadata."""

    STRING = "string"
    INT = "int"
    FLOAT = "float"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    # These two are only used by pydantic to select the model to parse.
    # The pydantic models for the metadata columns never set their type to either one of these values.
    CAT_INT = "categorical_int"
    CAT_STRING = "categorical_string"


CATEGORICAL_TYPE_PREFIX = "categorical_"


class Precision(IntEnum):
    """Precision of integer and float data."""

    SINGLE = 32
    DOUBLE = 64


# Config / Dataset Connectors
# -----------------------------------------------------------------------------


class TimeAttackMethod(StrEnum):
    """Possible methods against timing attacks."""

    JITTER = "jitter"
    STALL = "stall"


# Private Databases
class PrivateDatabaseType(StrEnum):
    """Type of Private Database for the private data."""

    PATH = "PATH_DB"
    S3 = "S3_DB"


class AuthenticationType(StrEnum):
    """Type of Authenticator to identify users."""

    FREE_PASS = "free_pass"
    OIDC = "oidc"


# Logging
# -----------------------------------------------------------------------------


def init_logging(name: str = "root", level: str = "INFO") -> logging.Logger:
    # Set root logger config
    logging.basicConfig(
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)],
    )
    # Set level for current logger
    logging.getLogger(name).setLevel(level)

    for loggers in ["aio_pika", "aiormq", "botocore", "faker", "urllib3", "httpx"]:
        logging.getLogger(loggers).setLevel(logging.INFO)
    return logging.getLogger(name)


# Exceptions
# -----------------------------------------------------------------------------


class ExceptionType(StrEnum):
    """Lomas server exception types.

    To be used as discriminator when parsing corresponding models
    """

    INVALID_QUERY = "InvalidQueryException"
    EXTERNAL_LIBRARY = "ExternalLibraryException"
    UNAUTHORIZED_ACCESS = "UnauthorizedAccessException"
    INTERNAL_SERVER = "InternalServerException"



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from lomas_core.constants import DPLibraries
from lomas_core.models.constants import ExceptionType


class LomasServerExceptionModel(BaseModel):
    """Base model for lomas server exceptions."""

    model_config = ConfigDict(use_attribute_docstrings=True)
    type: str
    """Exception type."""


class InvalidQueryExceptionModel(LomasServerExceptionModel):
    """Exception directly related to the query.

    For example if it does not contain a DP mechanism or there is not enough DP budget.
    """

    type: Literal[ExceptionType.INVALID_QUERY] = ExceptionType.INVALID_QUERY
    """Exception type."""
    message: str
    """Exception error message.

    This is for exceptions directly related to the query.
    For example if it does not contain a DP mechanism or
    there is not enough DP budget.
    """
    # Note: we duplicate the class docstring to show it in the openapi doc.


class ExternalLibraryExceptionModel(LomasServerExceptionModel):
    """For exceptions from libraries external to the lomas packages."""

    type: Literal[ExceptionType.EXTERNAL_LIBRARY] = ExceptionType.EXTERNAL_LIBRARY
    """Exception type."""
    library: DPLibraries
    """The external library that caused the exception."""
    message: str
    """Exception error message.

    For exceptions from libraries external to the lomas packages.
    """


class UnauthorizedAccessExceptionModel(LomasServerExceptionModel):
    """Exception related to rights with regards to the query.

    (e.g. no user access for this dataset).
    """

    type: Literal[ExceptionType.UNAUTHORIZED_ACCESS] = ExceptionType.UNAUTHORIZED_ACCESS
    """Exception type."""
    message: str
    """Exception error message.

    Exception related to rights with regards to the query.
    (e.g. no user access for this dataset).
    """


class InternalServerExceptionModel(LomasServerExceptionModel):
    """For any unforseen internal exception."""

    type: Literal[ExceptionType.INTERNAL_SERVER] = ExceptionType.INTERNAL_SERVER
    """Exception type.

    For any unforseen internal exception.
    """


LomasServerExceptionType = Annotated[
    InvalidQueryExceptionModel
    | ExternalLibraryExceptionModel
    | UnauthorizedAccessExceptionModel
    | InternalServerExceptionModel,
    Field(discriminator="type"),
]

LomasServerExceptionTypeAdapter: TypeAdapter = TypeAdapter(LomasServerExceptionType)



import logging

from opentelemetry import metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from lomas_core.models.config import Telemetry


def init_traces_exporter(resource: Resource, telemetry_config: Telemetry) -> None:
    """
    Initializes the OpenTelemetry trace exporter with a given resource.

    Args:
        resource (Resource): The resource to associate with the trace telemetry.
    """
    exporter = OTLPSpanExporter(
        endpoint=str(telemetry_config.collector_endpoint), insecure=telemetry_config.collector_insecure
    )
    span_processor = BatchSpanProcessor(exporter)
    tracer_provider = TracerProvider(resource=resource)
    tracer_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(tracer_provider)


def init_metrics_exporter(resource: Resource, telemetry_config: Telemetry) -> None:
    """
    Initializes the OpenTelemetry metrics exporter with a given resource.

    Args:
        resource (Resource): The resource to associate with the metric telemetry.
    """
    exporter = OTLPMetricExporter(
        endpoint=str(telemetry_config.collector_endpoint), insecure=telemetry_config.collector_insecure
    )
    reader = PeriodicExportingMetricReader(exporter)
    meter_provider = MeterProvider(resource=resource, metric_readers=[reader])
    metrics.set_meter_provider(meter_provider)


def init_logs_exporter(resource: Resource, telemetry_config: Telemetry) -> None:
    """
    Initializes the OpenTelemetry logs exporter with a given resource.

    Args:
        resource (Resource): The resource to associate with the log telemetry.
    """
    exporter = OTLPLogExporter(
        endpoint=str(telemetry_config.collector_endpoint), insecure=telemetry_config.collector_insecure
    )
    logger_provider = LoggerProvider(resource=resource)
    set_logger_provider(logger_provider)
    logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))

    handler = LoggingHandler(level=logging.NOTSET, logger_provider=logger_provider)
    logging.getLogger().addHandler(handler)


def init_telemetry(telemetry_config: Telemetry) -> None:
    """
    Initializes all OpenTelemetry exporters with a shared resource.

    Args:
        resource (Resource): The resource to associate with the app and instance.
    """
    resource = Resource.create(
        {"service.name": telemetry_config.service_name, "service.instance.id": telemetry_config.service_id}
    )

    init_traces_exporter(resource, telemetry_config)
    init_metrics_exporter(resource, telemetry_config)
    init_logs_exporter(resource, telemetry_config)



from typing import Annotated, Self

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    UrlConstraints,
    model_validator,
)

from lomas_core.models.constants import (
    TimeAttackMethod,
)


class TimeAttack(BaseModel):
    """BaseModel for configs to prevent timing attacks."""

    method: TimeAttackMethod
    magnitude: float


class Telemetry(BaseModel):
    """Telemetry config."""

    enabled: bool
    service_name: Annotated[str, Field(default="lomas-server-app")]
    service_id: Annotated[str, Field(default="default-host")]
    collector_endpoint: Annotated[HttpUrl, UrlConstraints(default_port=4317)] | None = None
    collector_insecure: Annotated[bool, Field(default=False)]
    collector_log_correlation: Annotated[bool, Field(default=False)]

    @model_validator(mode="after")
    def options_set_if_enabled(self) -> Self:
        """Ensures that if enabled is set to True, all other fields are specified.

        Raises:
            ValueError: If any of the fields is not specified while enabled is True.
        """
        if self.enabled:
            missing = [k for k, v in dict(self).items() if v is None]
            if len(missing) > 0:
                raise ValueError(
                    f"If enabled set to True, all other fields must be specified. Missing fields: {missing}"
                )

        return self


class OIDCConfig(BaseModel):
    """Base model for oidc config returned from discovery endpoint."""

    # Only useful (to us) fields are present in the model.
    model_config = ConfigDict(extra="ignore")

    issuer: HttpUrl
    authorization_endpoint: HttpUrl
    token_endpoint: HttpUrl
    jwks_uri: HttpUrl
    userinfo_endpoint: HttpUrl
    device_authorization_endpoint: HttpUrl
    introspection_endpoint: HttpUrl


class OIDCDeviceCodeResponse(BaseModel):
    """Base model for oidc device code response."""

    model_config = ConfigDict(extra="ignore")

    user_code: str
    device_code: str
    verification_uri: str
    verification_uri_complete: str | None = None



import itertools
from base64 import b64decode
from typing import Any

import opendp.prelude as dp
import polars as pl

from lomas_core.error_handler import InvalidQueryException
from lomas_core.models.collections import Metadata
from lomas_core.models.requests import GetDummyContext, OpenDPQueryModel

dp.enable_features("contrib")


def build_margins_from_metadata(metadata: Metadata) -> list:
    """
    Build a list of Polars margins from dataset metadata.

    This function derives margin constraints at three levels:
    1. A global margin based on the total number of rows.
    2. Single-column margins using per-column partition length and cardinality.
    3. Multi-column grouping margins (up to 4 columns) by combining individual
       column constraints.

    Args:
        metadata (Metadata): The metadata model for the real dataset.

    Raises:
        ValueError:
            If the metadata does not define the total number of rows.

    Returns:
        list: A list of ``dp.polars.Margin`` objects encoding global, per-column,
            and multi-column grouping constraints inferred from the metadata.
    """
    margins = []
    # --------------------
    # Global margin
    # --------------------
    rows = metadata.rows
    if rows is None:
        raise ValueError("Metadata must contain 'rows'")

    # TODO: invariant should change depending of the metadata (nrows given or not)
    # TBD with new metadata structure
    margins.append(dp.polars.Margin(max_length=rows, invariant="lengths"))

    # --------------------
    # Column-level margins
    # --------------------
    columns = metadata.columns

    # Store constraints for grouping
    grouping_constraints = {}

    for column_name, col_meta in columns.items():
        max_partition_length = col_meta.max_partition_length
        cardinality = getattr(col_meta, "cardinality", None)

        by = [column_name]

        # Save constraints for later grouping
        grouping_constraints[column_name] = {
            "max_partition_length": max_partition_length,
            "cardinality": cardinality,
        }

        # Categorical columns
        margin_kwargs: dict[str, Any] = {
            "by": by,
            "invariant": "keys",
        }

        if max_partition_length is not None:
            margin_kwargs["max_length"] = max_partition_length

        if cardinality is not None:
            margin_kwargs["max_groups"] = cardinality

        margins.append(dp.polars.Margin(**margin_kwargs))

    # --------------------
    # Multi-column groupings
    # --------------------
    column_names = list(grouping_constraints.keys())

    # For now, no more than 4 combination
    for r in range(2, min(len(column_names) + 1, 5)):
        for combo in itertools.combinations(column_names, r):
            max_lengths = []
            max_groups = []

            for col in combo:
                c = grouping_constraints[col]

                if c["max_partition_length"] is not None:
                    max_lengths.append(c["max_partition_length"])

                if c["cardinality"] is not None:
                    max_groups.append(c["cardinality"])

            margin_kwargs = {
                "by": list(combo),
                "invariant": "keys",
            }

            # min(max_partition_length)
            max_lengths = [
                grouping_constraints[col]["max_partition_length"]
                for col in combo
                if grouping_constraints[col]["max_partition_length"] is not None
            ]

            if max_lengths:
                margin_kwargs["max_length"] = min(max_lengths)

            # max_groups logic: multiply max_groups of all columns from combo
            # If one group is None and other is not, we keep the max group from the other
            # if all none, max_groups is none.
            cardinalities = [
                grouping_constraints[col]["cardinality"]
                for col in combo
                if grouping_constraints[col]["cardinality"] is not None
            ]
            if all(c is not None for c in cardinalities):
                product = 1
                for c in cardinalities:  # type: ignore[assignment]
                    product *= c  # type: ignore[operator]
                margin_kwargs["max_groups"] = product

            margins.append(dp.polars.Margin(**margin_kwargs))

    return margins


def build_context(
    query_json: GetDummyContext, metadata: Metadata, margins: list, input_data: pl.LazyFrame
) -> dp.Context:
    """
    Construct a differential privacy context from query parameters and metadata.

    This function validates the provided privacy parameters and builds a
    ``dp.Context`` using either a Laplace (epsilon-based) or Gaussian (rho-based)
    privacy loss, depending on the input. Exactly one of ``epsilon`` or ``rho``
    must be specified.

    The context is created using a compositor with a fixed privacy budget split
    (currently set to 1) and the margins derived from the dataset metadata.

    Args:
        query_json (GetDummyContext): Request defining the privacy parameters, including \
            ``epsilon``, ``rho``, and optional ``delta``.
        metadata (Metadata): The metadata model for the real dataset.
        margins (list): List of ``dp.polars.Margin`` objects defining aggregation and\
            grouping constraints.
        input_data (pl.LazyFrame): Input dataset on which the differentially private context \
            will be applied.

    Raises:
        InvalidQueryExceptionModel:
            If both ``epsilon`` and ``rho`` are provided.
        InvalidQueryExceptionModel:
            If neither ``epsilon`` nor ``rho`` is provided.

    Returns:
        dp.Context: A dp context configured with\
            the requested privacy loss, margins, and input data.
    """
    epsilon = query_json.epsilon
    rho = query_json.rho
    delta = query_json.delta
    approx_zcdp = query_json.approx_zcdp

    if epsilon is not None and rho is not None:
        raise InvalidQueryException("Provide only one of epsilon or rho, not both.")

    if epsilon is None and rho is None:
        raise InvalidQueryException("One of epsilon or rho must be provided.")

    if epsilon:
        # Laplace
        return dp.Context.compositor(
            data=input_data,
            privacy_unit=dp.unit_of(contributions=metadata.max_ids),
            privacy_loss=dp.loss_of(
                epsilon=epsilon,
                delta=delta,
            ),
            split_evenly_over=1,  # fixed to 1 for now, spend whole budget sent by user
            margins=margins,
        )

    # Gaussian

    if approx_zcdp is False:
        # if approx_zcdp is false, delta is only used to compute the associated epsilon
        # user wants to perform zcdp and not approx-zcdp (rho given, delta None)
        delta = None
    return dp.Context.compositor(
        data=input_data,
        privacy_unit=dp.unit_of(contributions=metadata.max_ids),
        privacy_loss=dp.loss_of(
            rho=rho,
            delta=delta,
        ),
        split_evenly_over=1,
        margins=margins,
    )


def build_context_from_metadata(
    query_json: OpenDPQueryModel,
    metadata: Metadata,
    input_data: pl.LazyFrame,
) -> dp.Context:
    """
    Build a differential privacy context from the metadata.

    Args:
        query_json (OpenDPQueryModel): Query containing privacy parameters and \
            a serialized OpenDP Polars query plan.
        metadata (Metadata): The metadata model for the real dataset.
        input_data (pl.LazyFrame): Input dataset.

    Returns:
        dp.Context: The constructed differential privacy context.
    """
    margins = build_margins_from_metadata(metadata=metadata)
    return build_context(query_json, metadata, margins, input_data)


def deserialize_context_query(
    query_json: OpenDPQueryModel,
    metadata: Metadata,
    input_data: pl.LazyFrame,
) -> dp.polars.LazyFrameQuery:
    """
    Build a differential privacy context and deserialize a Polars query plan.

    Args:
        query_json (OpenDPQueryModel): Query containing privacy parameters and \
            a serialized OpenDP Polars query plan.
        metadata (Metadata): The metadata model for the real dataset.
        input_data (pl.LazyFrame): Input dataset.

    Returns:
        dp.polars.LazyFrameQuery: A deserialized query bound to a newly \
            constructed context.
    """
    context = build_context_from_metadata(query_json, metadata, input_data)
    serialized_plan = b64decode(query_json.opendp_json.encode("utf-8"))
    return context.deserialize_polars_plan(serialized_plan)



from typing import Annotated, Any, Literal
from uuid import UUID, uuid4

import pandas as pd
import polars as pl
from diffprivlib.validation import DiffprivlibMixin
from pydantic import (
    BaseModel,
    ConfigDict,
    Discriminator,
    Field,
    PlainSerializer,
    PlainValidator,
    ValidationInfo,
    field_validator,
)

from lomas_core.constants import DPLibraries
from lomas_core.models.exceptions import LomasServerExceptionType
from lomas_core.models.utils import (
    dataframe_from_dict,
    dataframe_to_dict,
    deserialize_model,
    polars_df_from_str,
    polars_df_to_str,
    serialize_model,
)


class ResponseModel(BaseModel):
    """Base model for any response from the server."""


class InitialBudgetResponse(ResponseModel):
    """Model for responses to initial budget queries."""

    initial_epsilon: float
    """The initial epsilon privacy loss budget."""
    initial_delta: float
    """The initial delta privacy loss budget."""


class SpentBudgetResponse(ResponseModel):
    """Model for responses to spent budget queries."""

    total_spent_epsilon: float
    """The total spent epsilon privacy loss budget."""
    total_spent_delta: float
    """The total spent delta privacy loss budget."""


class RemainingBudgetResponse(ResponseModel):
    """Model for responses to remaining budget queries."""

    remaining_epsilon: float
    """The remaining epsilon privacy loss budget."""
    remaining_delta: float
    """The remaining delta privacy loss budget."""


class DummyDsResponse(ResponseModel):
    """Model for responses to dummy dataset requests."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    dtypes: Any
    """The dummy_df column data types."""
    dummy_df: Annotated[pd.DataFrame, PlainSerializer(dataframe_to_dict), PlainValidator(dataframe_from_dict)]

    """The dummy dataframe."""

    @field_validator("dummy_df", mode="before")
    @classmethod
    def deserialize_dummy_df(cls, v: pd.DataFrame | dict, info: ValidationInfo) -> pd.DataFrame:
        """Decodes the dict representation of the dummy df with correct types.

        Only does so if the input value is not already a dataframe.
        Args:
            v (pd.DataFrame | dict): The dataframe to decode.
            info (ValidationInfo): Validation info to access other model fields.

        Returns:
            pd.DataFrame: The decoded dataframe.
        """
        if isinstance(v, pd.DataFrame):
            return v

        dtypes = info.data["dtypes"]
        dummy_df = dataframe_from_dict(v)
        dummy_df = dummy_df.astype(dtypes)
        return dummy_df


class CostResponse(ResponseModel):
    """Model for responses to cost estimation requests or queries."""

    model_config = ConfigDict(use_attribute_docstrings=True)

    epsilon: float
    """The epsilon cost of the query."""
    delta: float
    """The delta cost of the query."""


# Query Responses
# -----------------------------------------------------------------------------


# DiffPrivLib
class DiffPrivLibQueryResult(BaseModel):
    """Model for diffprivlib query result."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: Literal[DPLibraries.DIFFPRIVLIB] = DPLibraries.DIFFPRIVLIB
    """Result type description."""
    score: float
    """The trained model score."""
    model: Annotated[
        DiffprivlibMixin,
        PlainSerializer(serialize_model),
        PlainValidator(deserialize_model),
    ]
    """The trained model."""


# SmartnoiseSQL
class SmartnoiseSQLQueryResult(BaseModel):
    """Type for smartnoise_sql result type."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: Literal[DPLibraries.SMARTNOISE_SQL] = DPLibraries.SMARTNOISE_SQL
    """Result type description."""
    df: Annotated[
        pd.DataFrame,
        PlainSerializer(dataframe_to_dict),
        PlainValidator(dataframe_from_dict),
    ]
    """Dataframe containing the query result."""


# SmartnoiseSynth
class SmartnoiseSynthModel(BaseModel):
    """Type for smartnoise_synth result when it is a pickled model."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    """Result type description."""
    type: Literal[DPLibraries.SMARTNOISE_SYNTH] = DPLibraries.SMARTNOISE_SYNTH
    """Synthetic data generator model."""
    # model: Annotated[Synthesizer, PlainSerializer(serialize_model), PlainValidator(deserialize_model)]


class SmartnoiseSynthSamples(BaseModel):
    """Type for smartnoise_synth result when it is a dataframe of samples."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: Literal["sn_synth_samples"] = "sn_synth_samples"
    """Result type description."""
    df_samples: Annotated[
        pd.DataFrame,
        PlainSerializer(dataframe_to_dict),
        PlainValidator(dataframe_from_dict),
    ]
    """Dataframe containing the generated synthetic samples."""


# OpenDP
class OpenDPQueryResult(BaseModel):
    """Type for opendp result."""

    type: Literal[DPLibraries.OPENDP] = DPLibraries.OPENDP
    """Result type description."""
    value: int | float | list[int | float]
    """The result value of the query."""


class OpenDPPolarsQueryResult(BaseModel):
    """Type for opendp Polars result."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    type: Literal[DPLibraries.OPENDP_POLARS] = DPLibraries.OPENDP_POLARS
    """Result type description."""
    # order of PlainValidator and PlainSerializer matters in that case:
    # https://github.com/pydantic/pydantic/issues/8512
    value: Annotated[
        pl.DataFrame,
        PlainValidator(polars_df_from_str),
        PlainSerializer(polars_df_to_str),
    ]
    """The result value of the query."""


# Response object
QueryResultT = Annotated[
    DiffPrivLibQueryResult
    | SmartnoiseSQLQueryResult
    | SmartnoiseSynthModel
    | SmartnoiseSynthSamples
    | OpenDPQueryResult
    | OpenDPPolarsQueryResult,
    Discriminator("type"),
]


class QueryResponse(CostResponse):
    """Response to Lomas queries."""

    requested_by: str
    """The user that triggered the query."""
    result: QueryResultT
    """The query result object."""


class Job(ResponseModel):
    """Scheduled Job Response."""

    uid: UUID = Field(default_factory=uuid4)
    """Job unique identifier."""
    requested_by: str | None = None
    """Name of the user that requested this job."""
    status: Literal["in_progress", "failed", "complete"] = "in_progress"
    """Job status."""
    result: QueryResponse | CostResponse | None = None
    """Job result, if available."""
    error: LomasServerExceptionType | None = None
    """Job error, if any."""
    status_code: int = 200
    """Status code for job response."""



import json
import pickle
from base64 import b64decode, b64encode
from typing import Any

import pandas as pd
import polars as pl

PANDAS_SERIALIZATION_ORIENT = "tight"


def dataframe_to_dict(df: pd.DataFrame) -> dict:
    """Transforms pandas dataframe into a dictionary.

    Args:
        df (pd.DataFrame): The dataframe to "serialize".

    Returns:
        dict: The pandas dataframe in dictionary format.
    """
    return df.to_dict(orient=PANDAS_SERIALIZATION_ORIENT)


def dataframe_from_dict(serialized_df: pd.DataFrame | dict) -> pd.DataFrame:
    """Transforms input dict into pandas dataframe.

    If the input is already a dataframe, it is simply returned unmodified.

    Args:
        serialized_df (pd.DataFrame | dict): Dataframe in dict format.
            Or pd.Dataframe.

    Returns:
        pd.DataFrame: The transformed dataframe.
    """
    if isinstance(serialized_df, pd.DataFrame):
        return serialized_df

    return pd.DataFrame.from_dict(serialized_df, orient=PANDAS_SERIALIZATION_ORIENT)


def polars_df_to_str(df_pl: pl.DataFrame) -> str:
    """Convert a Polars DataFrame to a JSON string."""
    return df_pl.write_json()


def polars_df_from_str(serialized_pl: str | pl.DataFrame) -> pl.DataFrame:
    """Convert a Polars DataFrame from a JSON string."""
    if isinstance(serialized_pl, pl.DataFrame):
        return serialized_pl

    df_pl = json.loads(serialized_pl)
    return pl.DataFrame(df_pl)


def serialize_model(model: Any) -> str:
    """
    Serialise a python object into an utf-8 string.

    Fitted Smartnoise Synth synthesizer or fitted DiffPrivLib pipeline.

    Args:
        model (Any): An object to serialise

    Returns:
        str: string of serialised model
    """
    serialized = b64encode(pickle.dumps(model))
    return serialized.decode("utf-8")


def deserialize_model(serialized_model: Any) -> Any:
    """Deserialize a base64 encoded byte string into a python object.

    Args:
        serialized_model (Any): Encoded python object.

    Returns:
        Any: Deserialized python object.
    """
    if isinstance(serialized_model, str):
        raw_bytes = b64decode(serialized_model)
        return pickle.loads(raw_bytes)

    return serialized_model



from datetime import datetime
from pathlib import Path
from typing import Annotated, Any, Literal, Self

from pydantic import BaseModel, Discriminator, Field, HttpUrl, Tag, model_validator

from lomas_core.models.constants import (
    CARDINALITY_FIELD,
    CATEGORICAL_TYPE_PREFIX,
    DB_TYPE_FIELD,
    TYPE_FIELD,
    MetadataColumnType,
    Precision,
    PrivateDatabaseType,
)

# Dataset of User
# -----------------------------------------------------------------------------


class DatasetOfUser(BaseModel):
    """BaseModel for informations of a user on a dataset."""

    dataset_name: str
    initial_epsilon: float
    initial_delta: float
    total_spent_epsilon: float = Field(default=0.0)
    total_spent_delta: float = Field(default=0.0)


# User
# -----------------------------------------------------------------------------


class UserId(BaseModel):
    """BaseModel for user identification."""

    name: str
    email: str
    client_secret: Annotated[
        str | None,
        Field(default=None, exclude=True),  # exclude the field at serialization for security reasons
    ]


class User(BaseModel):
    """BaseModel for a user in a user collection."""

    id: UserId
    may_query: bool
    admin: bool = False
    datasets_list: list[DatasetOfUser]


class UserCollection(BaseModel):
    """BaseModel for users collection."""

    users: list[User]


# Dataset Access Data
# -----------------------------------------------------------------------------


class DSPathAccess(BaseModel):
    """BaseModel for a local dataset."""

    database_type: Literal[PrivateDatabaseType.PATH]
    path: HttpUrl | Path  # force check Path should be relative ? or move prefix logic closer


class DSS3Access(BaseModel):
    """BaseModel for a dataset on S3."""

    database_type: Literal[PrivateDatabaseType.S3]
    endpoint_url: HttpUrl
    bucket: str
    key: str
    access_key_id: str | None = None
    secret_access_key: str | None = None
    credentials_name: str


class DSInfo(BaseModel):
    """BaseModel for a dataset."""

    dataset_name: str
    dataset_access: Annotated[DSPathAccess | DSS3Access, Field(discriminator=DB_TYPE_FIELD)]
    metadata_access: Annotated[DSPathAccess | DSS3Access, Field(discriminator=DB_TYPE_FIELD)]


class DatasetsCollection(BaseModel):
    """BaseModel for datasets collection."""

    datasets: list[DSInfo]


# Metadata
# -----------------------------------------------------------------------------


class ColumnMetadata(BaseModel):
    """Base model for column metadata."""

    private_id: bool = False
    nullable_proportion: Annotated[float, Field(ge=0, lt=1)] = 0.0
    # See issue #323 for checking this and validating.

    max_partition_length: Annotated[int | None, Field(gt=0)] = None
    max_influenced_partitions: Annotated[int | None, Field(gt=0)] = None
    max_partition_contributions: Annotated[int | None, Field(gt=0)] = None


class StrMetadata(ColumnMetadata):
    """Model for string metadata."""

    type: Literal[MetadataColumnType.STRING]


class CategoricalColumnMetadata(ColumnMetadata):
    """Model for categorical column metadata."""

    @model_validator(mode="after")
    def validate_categories(self) -> Self:
        """Makes sure number of categories matches cardinality."""
        if len(self.categories) != self.cardinality:
            raise ValueError("Number of categories should be equal to cardinality.")
        return self


class StrCategoricalMetadata(CategoricalColumnMetadata):
    """Model for categorical string metadata."""

    # The type must be kept to string (NOT categorical_string).
    # Some functions rely on this attribute
    # (e.g. data_connector when building pandas dataframe).
    type: Literal[MetadataColumnType.STRING]
    cardinality: int
    categories: list[str]


class BoundedColumnMetadata(ColumnMetadata):
    """Model for columns with bounded data."""

    @model_validator(mode="after")
    def validate_bounds(self) -> Self:
        """Validates column bounds."""
        if self.lower is not None and self.upper is not None and self.lower > self.upper:
            raise ValueError("Lower bound cannot be larger than upper bound.")

        return self


class IntMetadata(BoundedColumnMetadata):
    """Model for integer column metadata."""

    type: Literal[MetadataColumnType.INT]
    precision: Precision
    lower: int
    upper: int


class IntCategoricalMetadata(CategoricalColumnMetadata):
    """Model for integer categorical column metadata."""

    # The type must be kept to int (NOT categorical_int).
    # Some functions rely on this attribute
    # (e.g. data_connector when building pandas dataframe).
    type: Literal[MetadataColumnType.INT]
    precision: Precision
    cardinality: int
    categories: list[int]


class FloatMetadata(BoundedColumnMetadata):
    """Model for float column metadata."""

    type: Literal[MetadataColumnType.FLOAT]
    precision: Precision
    lower: float
    upper: float
    int_with_nulls: bool = False


class BooleanMetadata(ColumnMetadata):
    """Model for boolean column metadata."""

    type: Literal[MetadataColumnType.BOOLEAN]


class DatetimeMetadata(BoundedColumnMetadata):
    """Model for datetime column metadata."""

    type: Literal[MetadataColumnType.DATETIME]
    lower: datetime
    upper: datetime


def get_column_metadata_discriminator(v: Any) -> str:
    """Discriminator function for determining the type of column metadata.

    Args:
        v (Any): The unparsed column metadata (either dict or class object)

    Raises:
        ValueError: If the column type cannot be found.

    Returns:
        str: The metadata string type.
    """
    if isinstance(v, dict):
        col_type = v.get(TYPE_FIELD)
    else:
        col_type = getattr(v, TYPE_FIELD)

    if (
        col_type
        in {
            MetadataColumnType.STRING,
            MetadataColumnType.INT,
        }
    ) and (((isinstance(v, dict)) and CARDINALITY_FIELD in v) or (hasattr(v, CARDINALITY_FIELD))):
        col_type = f"{CATEGORICAL_TYPE_PREFIX}{col_type}"

    if not isinstance(col_type, str):
        raise ValueError("Could not find column type.")

    return col_type


class Metadata(BaseModel):
    """
    BaseModel for a metadata format.

    See Smartnoise-SQL documentation https://docs.smartnoise.org/sql/metadata.html
    """

    max_ids: Annotated[int, Field(gt=0)]
    rows: Annotated[int, Field(gt=0)]
    row_privacy: bool
    censor_dims: bool = True
    clamp_counts: bool = True
    clamp_columns: bool = True
    use_dpsu: bool = False
    # When parsing input data, pydantic first calls the discriminator function with the input data.
    # The model to build is then selected by matching the returned discriminator
    # with the tag that annotates each possible model.
    # The discriminator function is used to differentiate between int and categorical_int
    # or string and categorical_string columns.
    # The integer and string models always keep their type to int or string (not categorical_**).
    columns: dict[
        str,
        Annotated[
            Annotated[StrMetadata, Tag(MetadataColumnType.STRING)]
            | Annotated[StrCategoricalMetadata, Tag(MetadataColumnType.CAT_STRING)]
            | Annotated[IntMetadata, Tag(MetadataColumnType.INT)]
            | Annotated[IntCategoricalMetadata, Tag(MetadataColumnType.CAT_INT)]
            | Annotated[FloatMetadata, Tag(MetadataColumnType.FLOAT)]
            | Annotated[BooleanMetadata, Tag(MetadataColumnType.BOOLEAN)]
            | Annotated[DatetimeMetadata, Tag(MetadataColumnType.DATETIME)],
            Discriminator(get_column_metadata_discriminator),
        ],
    ]



from base64 import b64encode

import opendp.prelude as dp  # noqa: F401
import polars as pl
from pydantic import JsonValue

from lomas_core.constants import (
    SSynthGanSynthesizer,
)
from lomas_core.models.constants import (
    DIFFPRIVLIB_VERSION,
    DUMMY_NB_ROWS,
    DUMMY_SEED,
    OPENDP_VERSION,
)

# Query constants
PENGUIN_DATASET: str = "PENGUIN"
FSO_INCOME_DATASET: str = "FSO_INCOME_SYNTHETIC"
COVID_DATASET: str = "COVID_SYNTHETIC"
QUERY_EPSILON: float = 0.1
QUERY_DELTA: float = 0.00001
SQL_QUERY: str = "SELECT COUNT(*) AS NB_ROW FROM df"
DP_MECHANISM: JsonValue = {"count": "gaussian"}
FEATURE_COLUMNS: JsonValue = [
    "bill_length_mm",
    "bill_depth_mm",
    "flipper_length_mm",
    "body_mass_g",
]
TARGET_COLUMNS: JsonValue = ["species"]
SPLIT_SEED: int = 4
TEST_SIZE: float = 0.2
IMPUTER_STRATEGY: str = "drop"
SNSYNTH_NB_SAMPLES: int = 200


def make_dummy(example_query: dict[str, JsonValue]) -> dict[str, JsonValue]:
    """Make dummy example dummy query based on example query."""
    example_query_dummy = dict(example_query)
    example_query_dummy["dummy_nb_rows"] = DUMMY_NB_ROWS
    example_query_dummy["dummy_seed"] = DUMMY_SEED
    return example_query_dummy


# Lomas logic
# -----------------------------------------------------------------------------

example_get_admin_db_data: dict[str, JsonValue] = {
    "dataset_name": PENGUIN_DATASET,
}

example_get_dummy_dataset: dict[str, JsonValue] = {
    "dataset_name": PENGUIN_DATASET,
    "dummy_nb_rows": DUMMY_NB_ROWS,
    "dummy_seed": DUMMY_SEED,
}

# Smartnoise-SQL
# -----------------------------------------------------------------------------

example_smartnoise_sql_cost: dict[str, JsonValue] = {
    "query_str": SQL_QUERY,
    "dataset_name": PENGUIN_DATASET,
    "epsilon": QUERY_EPSILON,
    "delta": QUERY_DELTA,
    "mechanisms": DP_MECHANISM,
}

example_smartnoise_sql: dict[str, JsonValue] = dict(example_smartnoise_sql_cost)
example_smartnoise_sql["postprocess"] = True

example_dummy_smartnoise_sql: dict[str, JsonValue] = make_dummy(example_smartnoise_sql)

# Smartnoise-Synth
# -----------------------------------------------------------------------------

example_smartnoise_synth_cost: dict[str, JsonValue] = {
    "dataset_name": PENGUIN_DATASET,
    "synth_name": SSynthGanSynthesizer.DP_CTGAN,
    "epsilon": QUERY_EPSILON,
    "delta": QUERY_DELTA,
    "select_cols": [],
    "synth_params": {
        "embedding_dim": 128,
        "batch_size": 50,
        "epochs": 5,
    },
    "nullable": True,
    "constraints": "",
}
example_smartnoise_synth_query: dict[str, JsonValue] = dict(example_smartnoise_synth_cost)
example_smartnoise_synth_query["return_model"] = True
example_smartnoise_synth_query["condition"] = ""
example_smartnoise_synth_query["nb_samples"] = SNSYNTH_NB_SAMPLES

example_dummy_smartnoise_synth_query: dict[str, JsonValue] = make_dummy(example_smartnoise_synth_query)

# OpenDP
# -----------------------------------------------------------------------------

OPENDP_PIPELINE: str = (
    f'{{"version": "{OPENDP_VERSION}", '
    '"ast": {'
    '"_type": "partial_chain", "lhs": {'
    '"_type": "partial_chain", "lhs": {'
    '"_type": "partial_chain", "lhs": {'
    '"_type": "partial_chain", "lhs": {'
    '"_type": "partial_chain", "lhs": {'
    '"_type": "constructor", '
    '"func": "make_chain_tt", '
    '"module": "combinators", '
    '"args": ['
    "{"
    '"_type": "constructor", '
    '"func": "make_select_column", '
    '"module": "transformations", '
    '"kwargs": {"key": "bill_length_mm", "TOA": "String"}'
    "}, {"
    '"_type": "constructor", '
    '"func": "make_split_dataframe", '
    '"module": "transformations", '
    '"kwargs": {"separator": ",", "col_names": {"_type": '
    '"list", "_items": ["species", "island", '
    '"bill_length_mm", "bill_depth_mm", "flipper_length_'
    'mm", "body_mass_g", "sex"]}}'
    "}]}, "
    '"rhs": {'
    '"_type": "constructor", '
    '"func": "then_cast_default", '
    '"module": "transformations", '
    '"kwargs": {"TOA": "f64"}'
    "}}, "
    '"rhs": {'
    '"_type": "constructor", '
    '"func": "then_clamp", '
    '"module": "transformations", '
    '"kwargs": {"bounds": [30.0, 65.0]}'
    "}}, "
    '"rhs": {'
    '"_type": "constructor", '
    '"func": "then_resize", '
    '"module": "transformations", '
    '"kwargs": {"size": 346, "constant": 43.61}'
    "}}, "
    '"rhs": {'
    '"_type": "constructor", '
    '"func": "then_variance", '
    '"module": "transformations"'
    "}}, "
    '"rhs": {'
    '"_type": "constructor", '
    '"func": "then_laplace", '
    '"module": "measurements", '
    '"kwargs": {"scale": 5.0}'
    "}}}"
)

example_opendp: dict[str, JsonValue] = {
    "dataset_name": PENGUIN_DATASET,
    "opendp_json": b64encode(OPENDP_PIPELINE.encode("utf-8")).decode("utf-8"),
    "epsilon": QUERY_EPSILON,
    "delta": QUERY_DELTA,
    "rho": None,
    "approx_zcdp": False,
}
example_dummy_opendp: dict[str, JsonValue] = make_dummy(example_opendp)

# OpenDP Polars
# -----------------------------------------------------------------------------
OPENDP_POLARS_PIPELINE_DICTS: list[dict] = [
    {
        "region": 1,
        "eco_branch": 85,
        "profession": 52,
        "education": 7,
        "age": 60,
        "sex": 1,
        "income": 23496.63345669291,
    },
    {
        "region": 6,
        "eco_branch": 16,
        "profession": 94,
        "education": 5,
        "age": 44,
        "sex": 0,
        "income": 55903.89391456765,
    },
    {
        "region": 5,
        "eco_branch": 71,
        "profession": 73,
        "education": 2,
        "age": 22,
        "sex": 1,
        "income": 7317.908354313357,
    },
    {
        "region": 4,
        "eco_branch": 25,
        "profession": 74,
        "education": 7,
        "age": 112,
        "sex": 1,
        "income": 82935.48602726562,
    },
    {
        "region": 4,
        "eco_branch": 16,
        "profession": 73,
        "education": 4,
        "age": 94,
        "sex": 0,
        "income": 63534.775513084416,
    },
]

OPENDP_POLARS_PIPELINE: bytes = pl.from_dicts(OPENDP_POLARS_PIPELINE_DICTS).lazy().serialize()

OPENDP_POLARS_PIPELINE_COVID_DICTS: list[dict] = [
    {
        "patient_id": 7013,
        "id": 1023,
        "date": "2022-08-06",
        "temporal": 4,
        "georegion": "BS",
        "agegroup": "70 - 79",
        "sex": "other",
        "testType": "rapid_antigen_test",
        "testResult": "other",
        "country": "other",
        "subType": "BA.2.75",
        "hospitalization": False,
        "death": True,
    },
    {
        "patient_id": 2739,
        "id": 540,
        "date": "2023-09-06",
        "temporal": 1,
        "georegion": "VS",
        "agegroup": "unknown",
        "sex": "other",
        "testType": "rapid_antigen_test",
        "testResult": "other",
        "country": "unknown",
        "subType": "XBB",
        "hospitalization": True,
        "death": False,
    },
]

OPENDP_POLARS_PIPELINE_COVID: bytes = pl.from_dicts(OPENDP_POLARS_PIPELINE_COVID_DICTS).lazy().serialize()

example_opendp_polars: dict[str, JsonValue] = {
    "dataset_name": FSO_INCOME_DATASET,
    "opendp_json": b64encode(OPENDP_POLARS_PIPELINE).decode("utf-8"),
    "epsilon": QUERY_EPSILON,
    "delta": QUERY_DELTA,
    "rho": None,
    "approx_zcdp": False,
}


example_opendp_polars_plan = {**example_opendp_polars}
example_opendp_polars_plan["opendp_json"] = b64encode(
    pl.from_dicts(OPENDP_POLARS_PIPELINE_DICTS)
    .lazy()
    .select(pl.col("income").fill_null(0).dp.mean(bounds=(1000, 100000), scale=100_000.0))
    .serialize()
).decode("utf-8")

example_opendp_polars_datetime: dict[str, JsonValue] = {
    "dataset_name": COVID_DATASET,
    "opendp_json": b64encode(OPENDP_POLARS_PIPELINE_COVID).decode("utf-8"),
    "epsilon": QUERY_EPSILON,
    "delta": QUERY_DELTA,
    "rho": None,
    "approx_zcdp": False,
}


# DiffPrivLib
# -----------------------------------------------------------------------------

DIFFPRIVLIB_PIPELINE: str = (
    '{"module": "diffprivlib", '
    f'"version": "{DIFFPRIVLIB_VERSION}", '
    '"pipeline": ['
    "{"
    '"type": "_dpl_type:StandardScaler", '
    '"name": "scaler", '
    '"params": {'
    '"with_mean": true, '
    '"with_std": true, '
    '"copy": true, '
    '"epsilon": 0.5, '
    '"bounds": {'
    '"_tuple": true, '
    '"_items": [[30.0, 13.0, 150.0, 2000.0], [65.0, 23.0, 250.0, 7000.0]]'
    "}, "
    '"random_state": null, '
    '"accountant": "_dpl_instance:BudgetAccountant"'
    "}"
    "}, "
    "{"
    '"type": "_dpl_type:LogisticRegression", '
    '"name": "classifier", '
    '"params": {'
    '"tol": 0.0001, '
    '"C": 1.0, '
    '"fit_intercept": true, '
    '"random_state": null, '
    '"max_iter": 100, '
    '"verbose": 0, '
    '"warm_start": false, '
    '"n_jobs": null, '
    '"epsilon": 1.0, '
    '"data_norm": 83.69469642643347, '
    '"accountant": "_dpl_instance:BudgetAccountant"'
    "}"
    "}"
    "]"
    "}"
)

example_diffprivlib: dict[str, JsonValue] = {
    "dataset_name": PENGUIN_DATASET,
    "diffprivlib_json": DIFFPRIVLIB_PIPELINE,
    "feature_columns": FEATURE_COLUMNS,
    "target_columns": TARGET_COLUMNS,
    "test_size": TEST_SIZE,
    "test_train_split_seed": SPLIT_SEED,
    "imputer_strategy": IMPUTER_STRATEGY,
}
example_dummy_diffprivlib: dict[str, JsonValue] = make_dummy(example_diffprivlib)
